package com.goldencode.p2j.persist;

import antlr.RecognitionException;
import antlr.TokenStreamException;
import com.goldencode.p2j.persist.hql.HQLAst;
import com.goldencode.p2j.persist.hql.HQLLexer;
import com.goldencode.p2j.persist.hql.HQLParser;
import com.goldencode.p2j.persist.hql.HQLParserTokenTypes;
import org.hibernate.QueryException;
import org.hibernate.engine.SessionFactoryImplementor;

import java.io.StringReader;
import java.util.Iterator;
import java.util.List;

public abstract class CommonSafeCompareFunction extends SafeCompareFunction implements HQLParserTokenTypes
{
   /** Data value constant for true */
   protected static final String VAL_TRUE = "true";

   /** Data value constant for false */
   protected static final String VAL_FALSE = "false";

   /** Data value constant for null */
   protected static final String VAL_NULL = "null";

   private static enum ArgType {NULL, TRUE, FALSE, NOT_NULLABLE, NULLABLE};


   protected abstract String equalRender();

   protected abstract String notEqualRender();

   protected abstract String falseNullableRender(String argStr);

   protected abstract String trueNullableRender(String arg);

   protected abstract String nullNullableRender(String arg);

   protected abstract String trueNotNullableRender(String arg);

   protected abstract String falseNotNullableRender(String arg);

   protected abstract String bothNotNullableRender(String arg1, String arg2);

   protected abstract String firstNotNullableRender(String arg1, String arg2);

   protected abstract String bothNullableRender(String arg1, String arg2);

   public final String render(List args, SessionFactoryImplementor factory) throws QueryException
   {
      if (args.size() != 2) {
         throw new QueryException("Function requires two arguments");
      }
      String[] argStrs = {(String)args.get(0), (String)args.get(1)};
      SQLFunctionArgument[] arguments = {new SQLFunctionArgument(argStrs[0]), new SQLFunctionArgument(argStrs[1])};
      ArgType[] types = {arguments[0].getArgType(), arguments[1].getArgType()};
      String result = null;
      switch (types[0])
      {
         case NULL:
            switch (types[1])
            {
               case NULL:
                  result = equalRender();
                  break;
               case TRUE:
               case FALSE:
               case NOT_NULLABLE:
                  result = notEqualRender();
                  break;
               case NULLABLE:
                  result = nullNullableRender(argStrs[1]);
                  break;
            }
            break;
         case TRUE:
            switch (types[1])
            {
               case TRUE:
                  result = equalRender();
                  break;
               case NULL:
               case FALSE:
                  result = notEqualRender();
                  break;
               case NOT_NULLABLE:
                  result = trueNotNullableRender(argStrs[1]);
                  break;
               case NULLABLE:
                  result = trueNullableRender(argStrs[1]);
                  break;
            }
            break;
         case FALSE:
            switch (types[1])
            {
               case FALSE:
                  result = equalRender();
                  break;
               case NULL:
               case TRUE:
                  result = notEqualRender();
                  break;
               case NOT_NULLABLE:
                  result = falseNotNullableRender(argStrs[1]);
                  break;
               case NULLABLE:
                  result = falseNullableRender(argStrs[1]);
                  break;
            }
            break;
         case NOT_NULLABLE:
            switch (types[1])
            {
               case NULL:
                  result = notEqualRender();
                  break;
               case TRUE:
                  result = trueNotNullableRender(argStrs[0]);
                  break;
               case FALSE:
                  result = falseNotNullableRender(argStrs[0]);
                  break;
               case NOT_NULLABLE:
                  result = bothNotNullableRender(argStrs[0], argStrs[1]);
                  break;
               case NULLABLE:
                  result = firstNotNullableRender(argStrs[0], argStrs[1]);
                  break;
            }
            break;
         case NULLABLE:
            switch (types[1])
            {
               case NULL:
                  result = nullNullableRender(argStrs[0]);
                  break;
               case TRUE:
                  result = trueNullableRender(argStrs[0]);
                  break;
               case FALSE:
                  result = falseNullableRender(argStrs[0]);
                  break;
               case NOT_NULLABLE:
                  result = firstNotNullableRender(argStrs[1], argStrs[0]);
                  break;
               case NULLABLE:
                  result = bothNullableRender(argStrs[0], argStrs[1]);
                  break;
            }
            break;
      }
      return result;
   }

   private static class SQLFunctionArgument
   {
      private String arg;
      private ArgType argType;
      private HQLAst hqlAst = null;

      private SQLFunctionArgument(String arg)
      {
         this.arg = arg;
         if (arg.equals(VAL_NULL))
         {
            argType = ArgType.NULL;
         }
         else if (arg.equals(VAL_TRUE))
         {
            argType = ArgType.TRUE;
         }
         else if (arg.equals(VAL_FALSE))
         {
            argType = ArgType.FALSE;
         }
         else
         {
            hqlAst = parseArgument();
            argType = isNotNullable()? ArgType.NOT_NULLABLE : ArgType.NULLABLE;
         }
      }

      private HQLAst parseArgument() throws QueryException
      {
         StringReader reader = new StringReader(arg);
         HQLLexer lexer = new HQLLexer(reader);
         HQLParser parser = new HQLParser(lexer);
         try {
            parser.expression();
         } catch (RecognitionException e) {
            throw new QueryException(e);
         } catch (TokenStreamException e) {
            throw new QueryException(e);
         }
         HQLAst hqlAst = (HQLAst) parser.getAST();
         hqlAst.fixups(null, null);
         return hqlAst;
      }

      private boolean isNotNullable()
      {
         if (hqlAst.getType() == ALIAS)
         {
            String alias = hqlAst.getText();
            HQLAst property = (HQLAst) hqlAst.getChildAt(0);

            // Does the alias have a property child?
            if (property != null && property.getType() == PROPERTY)
            {
               // Look up DMO implementation class which corresponds with
               // alias.
               RecordBuffer buffer = BufferManager.get().lookupBuffer(alias);
               Database database = buffer.getDatabase();
               Class<?> dmoClass = buffer.getDMOImplementationClass();
               Iterator<String> nnIter = DatabaseManager.notNullProperties(database,
                       dmoClass);
               String propertyName = property.getText();
               // TODO: it does not work so far because mapping is needed between aliases
               while (nnIter.hasNext())
               {
                  if (propertyName.equals(nnIter.next()))
                  {
                     return true;
                  }
               }
            }
         }
         return false;
      }

      public ArgType getArgType()
      {
         return argType;
      }

      public HQLAst getHqlAst()
      {
         return hqlAst;
      }
   }

}
